#!/usr/bin/perl
# dh.pl - By Onigumo - 29/07/2003.
# Contact: onigumo@free.fr
#
# This program is released under the GPL, see license file in same directory.
#
# ENJOY!

use LWP::UserAgent;
use HTTP::Request;
use HTTP::Response;
use HTTP::Cookies;
use Getopt::Std;

$version = "0.91";

#flush STDOUT buffer after each command
select(STDOUT);
$| = 1;

getopts("h:d:l:u:f:s:nFD");

$def_domino_files = "Domino_files.txt";
$commands_docs_dangerous = "Commands_Documents_DANGEROUS.txt";
$commands_docs = "Commands_Documents.txt";
$commands_views = "Commands_Views.txt";
$commands_nsf = "Commands_NSF.txt";

if (!$opt_h || ($opt_d && $opt_d !~ /.+\.nsf$/) || ($opt_u && $opt_u !~ /.+\:.+/) || ($opt_d && $opt_f))  {
   print "\n\nUSAGE: $0 [options...]\n";
   print "\t-h Host/IP (mandatory)\n";
   print "\t-d Remote NSF file | -f NSF files list (defaults to $def_domino_files)\n";
   print "\t-l logfile\n";
   print "\t-u Username:Password\n";
   print "\t-s request latency in ms\n";
   print "\t-F: fast ACL enumeration, only retry unsuccessful attempts\n";
   print "\t        * MIGHT GIVE FALSE POSITIVES *\n";
   print "\t-n: no full scan, only test for NSF database availability\n";
   print "\t-D: enable dangerous doc commands (like ?DeleteDocument)\n";
   print "\t        * HEAVILY DISCOURAGED *\n";
   print "\t    (don't come and cry if you deleted all your files...)\n\n";

   exit 1;
}

print "\n\nDomino Hunter v$version - By Onigumo (c)2003 (onigumo@free.fr)\n\n\n";
print "--[ DISCLAIMER :\n\n";
print "This software is provided \"as is\", without any guarantee about its working.\n";
print "In other words, anybody should feel responsible for his/her own stupidity.\n";
print "Keep in mind that Domino Hunter does not provide any IDS evasion techniques;\n";
print "its default behaviour is to identify itself as a security scanner in order to\n";
print "allow server admins to monitor the security scan. Domino Hunter is purposedly\n";
print "VERY noisy and produces huge quantities of easily trackable logs on the\n";
print "server.\n\n";
print "Please let me know if you write any hacks or extensions to this program.\n\n";

$host = $opt_h;
$database = $opt_d;
$logfile = $opt_l;

if ($opt_f) {
   $files = $opt_f;
}else {
   $files = $def_domino_files;
}

if ($opt_s) {
    $latency = $opt_s;
}

$authentication_failure = "error";
$browser_string = "DominoHunter /$version (security scanner)";
#$browser_string = "Mozilla/4.76 [en] (Win98; U)";

$browser = LWP::UserAgent->new;
$browser->agent($browser_string);
$browser->cookie_jar(HTTP::Cookies->new(file => 'cookies.txt', autosave => 1, ignore_discard => 1));
$browser->timeout(5);

if ($opt_u) {
   $opt_u =~ s/ /%20/g; $opt_u =~ /^(.+)\:(.+)$/;
   $username = $1; $password = $2;
   $login_form = "names.nsf/?Login";

   print "--[ Using credentials: $username / $password\n";
}

open (LOGFILE,"> $logfile") if ($logfile);

if (!$database) {
    ScanFiles($browser,$host,$files);
}
else {
    ScanSingleFile($browser,$host,$database);
}

my_print("\n--[ Finished.\n\n");

close (LOGFILE) if ($logfile);

sub ScanFiles {
        my($browser,$host,$files) = @_;
        my $retvalue, $output;

        my_print("\n--[ Scanning NSF list: $files...\n");

        open (FILE, "$files") or die "\n\nCan't Open $files \n\n$!";

        foreach $cgi (<FILE>) {
                chomp $cgi;
                my_print("\t-- Now scanning: [$cgi]\n\n") if (!$opt_n);
                ScanSingleFile($browser,$host,$cgi);
        }

        close(FILE);
}

sub ScanSingleFile {
    my($browser,$host,$database) = @_;
    my $retvalue,$url,$output,@views;

    ($retvalue,$url,$output) = TestDatabase($browser,$host,$database);

    if (!$opt_n && $retvalue) {
        OpenDatabase($browser,$host,$database);
        @views = EnumerateViews($browser,$host,$database); die "Error: no valid view found" if (!@views);
        EnumerateDocs($browser,$host,$database,@views);
        Summary();
    } else {
        if (!$opt_n || ($output !~ /^404/ && $output !~ /^500/)) {
           my_print("$url -> $output\n");
        }
    }
}


sub TestDatabase {
    my($browser,$host,$database) = @_;
    my $retvalue,$output,$content,$raw;

    $url = "http://$host/$database";
    ($output,$content,$raw) = GetUrl($browser,$url);

    $retvalue = ($output =~ /^200/);
    return ($retvalue,$url,$output);
}

sub OpenDatabase {
    my($browser,$host,$database) = @_;
    my @C_MSF, $output, $url, $content, $raw;

    open (F3, "$commands_nsf") or die "\n\nCan't open $commands_nsf \n\n$!";
    @C_NSF = <F3>; chop(@C_NSF); close(F3);

    my_print("\n--[ Trying default NSF commands on $database...\n\n");

    foreach $nsf (@C_NSF) {
        $url = "http://$host/$database$nsf";
        ($output,$content, $raw) = GetUrl($browser,$url);
        Output($url,$output);
    }
}

sub EnumerateViews {
    my($browser,$host,$database) = @_;
    my @C_Views, @Valid_Views, $output, $i, $_url, $url, $content, $raw, $viewid, $p;
    my @boundary = (0x11a, 4095);

    open (F2, "$commands_views") or die "\n\nCan't open $commands_views \n\n$!";
    @C_Views = <F2>; chop(@C_Views); close(F2);

    my_print("\n--[ Scanning available views...\n\n");

    for ($i = $boundary[0]; $i < $boundary[1]; $i += 4) {
        $viewid = sprintf "%03x",$i;
        $url = "http://$host/$database/$viewid";
        ($output,$content,$raw) = GetUrl($browser,$url);
        Output($url,$output);
        if ($output =~ /^200/) {
           push(@Valid_Views,$viewid);
           foreach $command (@C_Views) {
                   $_url = "$url$command&AutoFramed";
                   ($output,$content,$raw) = GetUrl($browser,$_url);
                   Output($_url,$output);
                   if ($content =~ /viewentries/) {
                        ParseViewEntries($content,sprintf("%03x",$i));
                   }
           }
        }
        $p = 100 * ($i + 4 - $boundary[0])/($boundary[1]-$boundary[0]);
        printf("--[%.0f\%]-\r",$p);
    }

    return @Valid_Views;
}

sub ParseViewEntries {
    my ($content, $vid) = @_;
    my @lines = split(/\n/,$content);

    foreach $line (@lines) {
            if ($line =~ /<viewentry[^>]*unid="([^"]*)"[^>]*noteid="([^"]*)"[^>]*>/) {
                $unid{$1} = $vid;
                $noteid{$2} = $vid;

                my_print("\t+ Document $1 / $2 officially in view $vid\n");
            }
    }
}

sub EnumerateDocs {
    my($browser,$host,$database,@views) = @_;
    my @C_Docs, $output, $i, $url, $content, $raw, $nid, $viewnoteid, @b, $offset, @tmp_access;

    #my @boundaries = (2294,4095,4150,6354,8390,10358,0,-1);
    my @boundaries = (2294,65534,0,-1);

    open (F1, "$commands_docs") or die "\n\nCan't open $commands_docs \n\n$!";
    @C_Docs = <F1>; chop(@C_Docs); close(F1);

    if ($opt_D) {
        my @tmplist;
        open (F2, "$commands_docs_DANGEROUS") or die "\n\nCan't open $commands_docs_DANGEROUS \n\n$!";
        @tmplist = <F2>; chop(@tmplist); close(F2);
        @C_Docs = (@C_Docs,@tmplist);
    }

    my_print("\n--[ Enumerating documents through valid views...");
    my_print(" * DANGEROUS COMMANDS ENABLED *") if ($opt_D);
    my_print("\n\n");

    foreach $viewid (@views) {
        my_print("\tCurrent view = $viewid\n\n");
        $offset = 0;
        @b = ($boundaries[$offset],$boundaries[$offset+1]);

        for ($i = $b[0]; $i < $b[1]; $i += 4) {
           next if ($opt_F && $tmp_access[$i] == 1);
           $nid = sprintf "%x",$i;
           $viewnoteid = $noteid{$nid};
           $url = "http://$host/$database/$viewid/$nid?OpenDocument&AutoFramed";
           ($output,$content,$raw) = GetUrl($browser,$url);
           if ($viewnoteid) {
                if ($viewid =~ /^$viewnoteid$/) {
                   Output("-- $url",$output);
                } else {
                   Output("++ $url",$output);
                }
           } else {
                Output(" ? $url",$output);
                $viewnoteid = $viewid;
           }

           if ($output =~ /^200/) {
              $tmp_access[$i] = 1;
              $access{$nid}[hex($viewnoteid)] = 1;
              foreach $doc_command (@C_Docs) {
                 my $o, $c, $r;
                 $url = "http://$host/$database/$viewid/$nid$doc_command";
                 ($o,$c,$r) = GetUrl($browser,$url);
                 Output("\t$viewid/$nid",$output);
              }
              ParseFile($nid,$content);
           } else {
              if ($viewid eq $viewnoteid) {
                 $access{$nid}[hex{$viewnoteid}] = -10;
              } else {
                 $access{$nid}[hex{$viewnoteid}] = -1;
              }
           }

           break if ($output =~ /^500/);

           if ($i+4 >= $b[1]) {
                $offset += 2;
                @b = ($boundaries[$offset],$boundaries[$offset+1]);
                $i = $b[0] - 4;
           }
        }
        my_print("\n");
    }
}

sub Summary {
        my $i,$docid,$restricted, $hexvalue, $file;
        my_print("\n--[ ACLs summary\n\n");

        foreach $docid (keys %access) {
           $restricted = 0;
           my_print("   doc = $docid>\t");

           for ($i = 282; $i < 4095; $i += 4) {
                if ($access{$docid}[$i] == 1) {
                    $hexvalue = sprintf "%03x",$i;
                    my_print("$hexvalue|");
                }
                elsif ($access{$docid}[$i] == -1) {
                    my_print("    |");
                }
                $restricted = ($access{$docid}[$i] == -10);
           }
           my_print(" (has documents)") if ($files_summary{$hexvalue} == 1);
           my_print(" (restricted on home view)") if ($restricted);
           my_print("\n");
        }
        my_print("\n--[ External files summary\n\n");
        foreach $file (keys %external_files) {
           my_print("   $file\n");
        }
}


sub GetUrl {
        my ($browser,$url) = @_;

        if ($opt_u) {
           if (!Login($browser,"http://$host/$login_form",$username,$password)) {
              die "Invalid credentials supplied\n\n$!";
           }
        }

        my $req = new HTTP::Request("GET" => "$url");
        my $response = $browser->request($req);
        my $content, $file, $shorturl;

        select(undef,undef,undef,$latency / 1000) if ($opt_n);

        return "Access denied" if ($response->status_line =~ /^200/ && $response->as_string =~ /$authentication_failure/);

        if ($response->status_line =~ /^200/) {
            $file = $url; $file =~ s/.*:\/\/[^\/]+\/([^\?]+).*/\/$1/;
            $shorturl = $file; $shorturl =~ s/^.*\.nsf(\/.*)$/$1/;
            $content = $response->content;
            $content =~ s#"$file[^"]*"#""#g;

            if (!$opt_F) {
               if (!$hash{$content}) {
                  $hash{$content} = "$shorturl";
               } else {
                  return ($response->status_line." (similar to: ".$hash{$content}.")",$content,$response->as_string);
              }
            }
        }

        return ($response->status_line,$content,$response->as_string);
}

sub Login {
        my ($browser,$url,$username,$password) = @_;

        return 1 if ($browser->cookie_jar->as_string() =~ /DomAuthSessId/);

        my $req = new HTTP::Request("POST" => "$url");
        $req->content_type('application/x-www-form-urlencoded');
        $req->content('Username='.$username.'&Password='.$password.'&%%ModDate=0000000000000000');
        my $response = $browser->request($req);

        if ($response->as_string !~ /$authentication_failure/) {
           my_print("\n--[ Login successful.\n\n");
           return 1;
        }

        return 0;
}


sub Output {
        my($url,$output) = @_;
        my $text;

        if ($output !~ /^404/ && $output !~ /^500/) {
           $text = "$url -> $output\n";
           my_print($text);
        }
}

sub my_print {
        my($arg) = @_;

        print $arg;
        print LOGFILE $arg if ($logfile);
}

sub ParseFile {
        my($docid,$content) = @_;
        my $file;

        if ($content =~ /\/\$FILE\//) {
           my_print("\t-- Doc $docid refers to external documents:\n");
           $files_summary{$docid} = 1;
           while ($content =~ s/<A HREF="[^"\$]*\/\$FILE\/([^"]*)">/<A HREF>/) {
                my_print("\t     $1\n");
                $external_files{$1} = 1;
           }
           my_print("\n");
        }
        else {
           $files_summary{$docid} = 0;
        }
}
